#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <getopt.h>
#include <pthread.h>
#include "cdflib.h"

#define max_motif_width 100
#define max_strbuffer_len 2500
#define max_seq_len 5500
#define alphsize 5
#define max_seq_n 50000
#define max_motif_n 5000
#define threads_n 8

int run_sampling = 0;
unsigned int random_seed = 12345;
int rdm_n = 2000;
double outlier_fc = 3.0;
double score_th = 1.0;
char DNA_Bases[5] = {'A', 'C', 'G', 'T', 'N'};
float pseudocount = 0.375;
pthread_mutex_t thread_lock;

struct motif {
    int len;
    char name[max_strbuffer_len];
    int matrix[max_motif_width][alphsize];
    float frqs[max_motif_width][alphsize];
    int copy_n;
};

struct seq {
    int len;
    char name[max_strbuffer_len];
    char seq[max_seq_len];
    short int base[max_seq_len];
    float background_frqs[alphsize];
    int start, end;
    char strand;
    char chrom[max_strbuffer_len];
};

struct match {
    int mi, si;
    double pvalue;
    double tvalue;
    double rs;
    double rank;
    char bsite[max_motif_width];
    int location, pos;
    int start, end;
    char strand;
    char chrom[max_strbuffer_len];
};

struct thread_para {
    int *motif_indexs;
    int msize;
    FILE *fpout;
};

struct motif *motif_db;
struct seq *seq_db;
//struct motif motif_db[max_motif_n];
//struct seq seq_db[max_seq_n];
float background_frqs[alphsize];
int motif_n = 0;
int seq_n = 0;
int finished_n = 0;

char const short_options[] = "hm:i:s:n:r:o:";
struct option long_options[] =
        {
            {"help", 0, NULL, 'h'},
            {"motif", 1, NULL, 'm'},
            {"input", 1, NULL, 'i'},
            {"seed", 1, NULL, 's'},
            {"rand_n", 1, NULL, 'n'},
            {"score", 1, NULL, 'r'},
            {"output", 1, NULL, 'o'},
            {0, 0, 0, 0}
        };

void rtrim(char *str)
{
    if(str == NULL || *str == '\0') {
        return;
    }
    int len = strlen(str);
    char *p = str + len - 1;
    while ((*p == '\r' || *p == '\n') && p >= str) {
        *p = '\0';
         --p;
    }
}

void normalize_pssm(struct motif *motif_inst_p) {
    int r = (*motif_inst_p).len;
    int c = alphsize;
    float tot;
    float tot_all = 0;
    for (int ri = 0; ri < r; ri++) {
        tot = 0;
        for (int ci = 0; ci < c; ci++) {
            (*motif_inst_p).frqs[ri][ci] = pseudocount;
        }
        for (int ci = 0; ci < c; ci++) {
            (*motif_inst_p).frqs[ri][ci] += (*motif_inst_p).matrix[ri][ci];
            tot += (*motif_inst_p).frqs[ri][ci];
        }
        tot_all += tot;
        for (int ci = 0; ci < c; ci++) {
            (*motif_inst_p).frqs[ri][ci] /= tot;
        }
    }
    (*motif_inst_p).copy_n = round(tot_all / r);
}

struct motif read_motif2(FILE *fpin) {
    struct motif motif_inst;
    char buffer[max_strbuffer_len];
    int matrix_len = 0;
    int header_ok = 0;
    fpos_t pos;
    do {
        fgetpos(fpin, &pos);
        if (fgets(buffer, max_strbuffer_len, fpin) == NULL) {
            break;
        }
        //scanf("%[^\n]\n", buffer);
        //fgets(buffer, max_strbuffer_len, fpin);
        rtrim(buffer);
        //printf("in: %s\n", buffer);
        if (buffer[0] == '>') {
            if (header_ok == 0) {
                char *p = buffer;
                memmove(buffer, p + 1, strlen(buffer));
                strcpy(motif_inst.name, buffer);
                header_ok = 1;
            }
            else {
                motif_inst.len = matrix_len;
                fsetpos(fpin, &pos);
                break;
            }
        }
        else {
            matrix_len++;
            //strcpy (motif_inst.matrix[matrix_len], buffer);
            //printf("row: %s\n", buffer);
            char *token;
            token = strtok(buffer, "\t"); 
            int ci = 0;
            while (token != NULL) {
                //printf("Token: %s\n", token);
                ci++;
                if (ci > max_motif_width) {
                    printf("motif width: >%d\n", max_motif_width);
                    break;
                }
                motif_inst.matrix[matrix_len - 1][ci - 1] = atoi(token);
                motif_inst.matrix[matrix_len - 1][alphsize - 1] = 0;
                token = strtok(NULL, "\t");
            }
        }
    } while(!feof(fpin));
    if (header_ok == 1) {
        motif_inst.len = matrix_len;
    }
    normalize_pssm(&motif_inst);
    return motif_inst;
}

int base_2int(char base) {
    if (base == 'A' || base == 'a') {
        return 0;
    }
    else if (base == 'C' || base == 'c') {
        return 1;
    }
    else if (base == 'G' || base == 'g') {
        return 2;
    }
    else if (base == 'T' || base == 't') {
        return 3;
    }
    else {
        return 4;
    }
}

void calculat_bgfrqs(struct seq *seq_inst_p) {
    int c = alphsize;
    int tot = 0;
    for (int ci = 0; ci < c; ci++) {
        (*seq_inst_p).background_frqs[ci] = pseudocount;
        tot += pseudocount;
    }
    for (int i=0; i<(*seq_inst_p).len; i++) {
        tot++;
        (*seq_inst_p).background_frqs[(*seq_inst_p).base[i]]++;
    }
    for (int ci = 0; ci < c; ci++) {
        (*seq_inst_p).background_frqs[ci] /= tot;
    }
}

void parse_seq_title(struct seq *seq_inst_p) {
    char *token;
    token = strtok((*seq_inst_p).name, "::");
    token = strtok(NULL, ":");
    strcpy((*seq_inst_p).chrom, token);
    token = strtok(NULL, "-");
    (*seq_inst_p).start = atoi(token);
    token = strtok(NULL, "(");
    (*seq_inst_p).end = atoi(token);
    token = strtok(NULL, ")");
    (*seq_inst_p).strand = *token;
}

struct seq read_seq2(FILE *fpin) {
    struct seq seq_inst;
    char buffer[max_strbuffer_len];
    char seq_buffer[max_seq_len];
    memset(seq_buffer, '\0', strlen(seq_buffer));
    int header_ok = 0;
    fpos_t pos;
    do {
        fgetpos(fpin, &pos);
        if (fgets(buffer, max_strbuffer_len, fpin) == NULL) {
            break;
        }
        //scanf("%[^\n]\n", buffer);
        rtrim(buffer);
        //printf("in: %s\n", buffer);
        if (buffer[0] == '>') {
            if (header_ok == 0) {
                char *p = buffer;
                memmove(buffer, p + 1, strlen(buffer));
                strcpy(seq_inst.name, buffer);
                header_ok = 1;
            }
            else {
                strcpy(seq_inst.seq, seq_buffer);
                seq_inst.len = strlen(seq_buffer);
                for (int i=0; i<seq_inst.len; i++) {
                    seq_inst.base[i] = base_2int(seq_buffer[i]);
                }
                fsetpos(fpin, &pos);
                break;
            }
        }
        else {
            if (strlen(seq_buffer) + strlen(buffer) > max_seq_len) {
                printf("seq len: >%d\n", max_seq_len);
                break;
            }
            else {
                strcat(seq_buffer, buffer);
            }
        }
    } while(!feof(fpin));
    calculat_bgfrqs(&seq_inst);
    parse_seq_title(&seq_inst);
    return seq_inst;
}

void calculated_pos2(int chr_start, int chr_end, char strand, struct match *match_inst_p) {
    if (strand == '+') {
        (*match_inst_p).start = chr_end + (*match_inst_p).pos + 1;
        (*match_inst_p).end = (*match_inst_p).start + strlen((*match_inst_p).bsite) - 1;
    } else if (strand == '-') {
        (*match_inst_p).end = chr_start - (*match_inst_p).pos - 1;
        (*match_inst_p).start = (*match_inst_p).end - strlen((*match_inst_p).bsite) + 1;
    } else {
        printf("unknown strand!\n");
        return;
    }
}

struct match scan_seq_match(int mi, int si) {
    struct seq *seq_inst_p = &seq_db[si];
    struct motif *motif_inst_p = &motif_db[mi];
    unsigned int pssm_len = (*motif_inst_p).len;
    long double max_rs = 0.0;
    int best_pos = -1;
    int posns = (*seq_inst_p).len - pssm_len + 1;
    for (int n = 0; n < posns; ++n) {
        long double rs_avg = 1.0;
        for (int r = 0; r < pssm_len; ++r) {
            int c = (*seq_inst_p).base[n+r];
            rs_avg *= (*motif_inst_p).frqs[r][c] / background_frqs[c];
        }
        if (max_rs < rs_avg) {
            max_rs = rs_avg;
            best_pos = n;
        }
    }
    struct match match_rt;
    match_rt.mi = mi;
    match_rt.si = si;
    match_rt.rs = log(pow(max_rs, (double)1.0 / pssm_len));
    match_rt.location = best_pos;
    match_rt.pos = match_rt.location - (*seq_inst_p).len;
    strncpy(match_rt.bsite, (*seq_inst_p).seq + best_pos, (*motif_inst_p).len);
    match_rt.bsite[(*motif_inst_p).len] = '\0';
    match_rt.strand = (*seq_inst_p).strand;
    strcpy(match_rt.chrom, (*seq_inst_p).chrom);
    calculated_pos2((*seq_inst_p).start, (*seq_inst_p).end, (*seq_inst_p).strand, &match_rt);
    return match_rt;
}

void std_mean(double *data, int n, double *std, double *meanr) {
    if (n < 2) {
        *std = 0.0;
        *meanr = 0.0;
        return;
    }
    double total = 0.0;
    double mean;
    for (int i = 0; i < n; i++) {
        total += data[i];
    }
    mean = total / (double) n;
    double cov_cum = 0.0;
    for (int i = 0; i < n; i++) {
        cov_cum += pow((data[i] - mean), 2.0);
    }
    //return cov_cum;
    *std = sqrt(cov_cum / (double)n);
    *meanr = mean;
}

void one_sample_ttest(double one, double *rdm_rs, int n, double *pvalue, double *t) {
    if (n < 3) {
        *pvalue = 1.0;
        *t = 0.0;
        return;
    }
    double bound_l, bound_r;
    double df;
    double p_l, p_r;
    double q_l, q_r;
    int status_l, status_r;
    double tt, ts, t_l, t_r;
    int which_l = 1, which_r = 1;
    //double t, df;
    //double tmp_ccv, tmp_sd;
    double std;
    double mean;
    double diff;
    std_mean(rdm_rs, n, &std, &mean);
    diff = one - mean;
    ts = (one - mean) / std;
    tt = (one - mean) / (std / sqrt(n));
    //tt = (one - mean) / std;
    df = n - 1.0;
    t_l = fabs(tt) * -1.0;
    cdft(&which_l, &p_l, &q_l, &t_l, &df, &status_l, &bound_l);
    t_r = fabs(tt);
    cdft(&which_r, &p_r, &q_r, &t_r, &df, &status_r, &bound_r);
    double p2;
    if (status_l != 0 || status_r != 0) {
        p2 = 1.0;
    } else {
        p2 = p_l + q_r;
    }
    *pvalue = p2;
    *t = ts;
}

void sampling_noputback(int n, int rand_n, int *index_rdm) {
    int *index_pool = malloc(sizeof(int) * n);
    for (int i = 0; i<n; i++) {
        index_pool[i] = i;
    }
    for (int i = 0; i<rand_n; i++) {
        int ret = rand()%(n - i);
        //printf("%d\n", index_pool_p[ret]);
        index_rdm[i] = index_pool[ret];
        index_pool[ret] = index_pool[n - i - 1];
    }
    free(index_pool);
}

void random_matchs(int mi, int rand_n, struct match *matchs, int *rt_n) {
    struct motif *motif_inst_p = &motif_db[mi];
    struct match match_one;
    struct match *matchs_tmp = malloc(sizeof(struct match) * rand_n);
    double *rss_tmp = malloc(sizeof(double) * rand_n);
    double *rss = malloc(sizeof(double) * rand_n);
    int *index_rdm = malloc(sizeof(int) * rand_n);
    sampling_noputback(seq_n, rand_n, index_rdm);
    int si;
    for (int i = 0; i < rand_n; i++) {
        //int si = rand() % seq_n;
        si = index_rdm[i];
        match_one = scan_seq_match(mi, si);
        matchs_tmp[i] = match_one;
        rss_tmp[i] = match_one.rs;
    }
    double std_dev;
    double mean;
    double diff;
    std_mean(rss_tmp, rand_n, &std_dev, &mean);
    int i;
    int rn = 0;
    for (i = 0; i < rand_n; i++) {
        if (fabs(rss_tmp[i] - mean) < outlier_fc * std_dev) {
            rn++;
            matchs[rn - 1] = matchs_tmp[i];
            rss[rn - 1] = rss_tmp[i];
        }
    }
    *rt_n = rn;
    double pvalue, tvalue;
    for (int i = 0; i < rn; i++) {
        one_sample_ttest(rss[i], rss, rn, &pvalue, &tvalue);
        matchs[i].pvalue = pvalue;
        matchs[i].tvalue = tvalue;
    }
    free(matchs_tmp);
    free(rss_tmp);
    free(rss);
    free(index_rdm);
}

void all_matchs(int mi, struct match *matchs, int *rt_n) {
    struct motif *motif_inst_p = &motif_db[mi];
    struct match match_one;
    struct match *matchs_tmp = malloc(sizeof(struct match) * seq_n);
    double *rss_tmp = malloc(sizeof(double) * seq_n);
    double *rss = malloc(sizeof(double) * seq_n);
    //int *index_rdm = malloc(sizeof(int) * rand_n);
    //sampling_noputback(seq_n, rand_n, index_rdm);
    int si;
    for (int si = 0; si < seq_n; si++) {
        //int si = rand() % seq_n;
        //si = index_rdm[i];
        match_one = scan_seq_match(mi, si);
        matchs_tmp[si] = match_one;
        rss_tmp[si] = match_one.rs;
    }
    double std_dev;
    double mean;
    double diff;
    std_mean(rss_tmp, seq_n, &std_dev, &mean);
    int i;
    int rn = 0;
    for (i = 0; i < seq_n; i++) {
        if (fabs(rss_tmp[i] - mean) < outlier_fc * std_dev) {
            rn++;
            matchs[rn - 1] = matchs_tmp[i];
            rss[rn - 1] = rss_tmp[i];
        }
    }
    *rt_n = rn;
    double pvalue, tvalue;
    for (int i = 0; i < rn; i++) {
        one_sample_ttest(rss[i], rss, rn, &pvalue, &tvalue);
        matchs[i].pvalue = pvalue;
        matchs[i].tvalue = tvalue;
    }
    free(matchs_tmp);
    free(rss_tmp);
    free(rss);
    //free(index_rdm);
}

void calculat_bgfrqs_all() {
    struct seq *seq_inst_p;
    int c = alphsize;
    int tot = 0;
    for (int ci = 0; ci < c; ci++) {
        background_frqs[ci] = pseudocount;
        tot += pseudocount;
    }
    for (int si = 0; si < seq_n; si++) {
        seq_inst_p = &seq_db[si];
        for (int i=0; i<(*seq_inst_p).len; i++) {
            tot++;
            background_frqs[(*seq_inst_p).base[i]]++;
        }
    }
    for (int ci = 0; ci < c; ci++) {
        background_frqs[ci] /= tot;
    }
}

void read_inputs(char *motif_file, char *seq_file) {
    FILE *fpin;
    fpin = fopen(motif_file, "rb");
    while (!feof(fpin)) {
        struct motif motif_inst;
        motif_inst = read_motif2(fpin);
        //printf("motif_name: %s\n", motif_inst.name);
        motif_n++;
        if (motif_n > max_motif_n) {
            printf("motif_n: >%d\n", max_motif_n);
            break;
        }
        motif_db[motif_n - 1] = motif_inst;
    }
    fclose(fpin);
    fpin = fopen(seq_file, "rb");
    while (!feof(fpin)) {
        struct seq seq_inst;
        seq_inst = read_seq2(fpin);
        //printf("motif_name: %s\n", seq_inst.name);
        seq_n++;
        if (seq_n > max_seq_n) {
            printf("seq_n: >%d\n", max_seq_n);
            break;
        }
        seq_db[seq_n - 1] = seq_inst;
    }
    fclose(fpin);
    calculat_bgfrqs_all();
}

void random_gene_set(char *motif_file, char *seq_file, char *output_file) {
    FILE *fpout;
    fpout = fopen(output_file, "w");
    read_inputs(motif_file, seq_file);
    struct match *match_db = malloc(sizeof(struct match) * rdm_n);
    //struct match matchs[20000];
    int matchs_n;
    for (int mi = 0; mi < motif_n; mi++) {
        printf("%d/%d\n", mi + 1, motif_n);
        random_matchs(mi, rdm_n, match_db, &matchs_n);
        for (int ri = 0; ri < matchs_n; ri++) {
            if (fabs(match_db[ri].tvalue) < score_th) {
                continue;
            }
            fprintf(fpout, "%s\t%s\t%s\t%d\t%d\t%c\t%s\t%f\t%f\t%e\n", motif_db[match_db[ri].mi].name, seq_db[match_db[ri].si].name, match_db[ri].chrom,
            match_db[ri].start, match_db[ri].end, match_db[ri].strand, match_db[ri].bsite, match_db[ri].rs, match_db[ri].tvalue, match_db[ri].pvalue);
        }
    }
    free(match_db);
    fclose(fpout);
}

void *extream_threading(void* args) {
    struct thread_para *targs = (struct thread_para*)args;
    //FILE *fpout;
    //fpout = fopen(output_file, "w");
    //read_inputs(motif_file, seq_file);
    struct match *match_db = malloc(sizeof(struct match) * seq_n);
    //struct match matchs[20000];
    int matchs_n;
    //for (int mi = 0; mi < motif_n; mi++) {
    for (int ix = 0; ix < (*targs).msize; ix++) {
        int mi = (*targs).motif_indexs[ix];
        all_matchs(mi, match_db, &matchs_n);
        for (int ri = 0; ri < matchs_n; ri++) {
            if (match_db[ri].tvalue < score_th) {
                continue;
            }
            pthread_mutex_lock(&thread_lock);
            fprintf((*targs).fpout, "%s\t%s\t%s\t%d\t%d\t%c\t%s\t%f\t%f\t%e\n", motif_db[match_db[ri].mi].name, seq_db[match_db[ri].si].name, match_db[ri].chrom,
            match_db[ri].start, match_db[ri].end, match_db[ri].strand, match_db[ri].bsite, match_db[ri].rs, match_db[ri].tvalue, match_db[ri].pvalue);
            pthread_mutex_unlock(&thread_lock);
        }
        finished_n++;
        pthread_mutex_lock(&thread_lock);
        printf("%d/%d\n", finished_n, motif_n);
        pthread_mutex_unlock(&thread_lock);
    }
    free(match_db);
    //fclose(fpout);
}

void assign_threads(struct thread_para *paras, FILE *fpout) {
    //printf("ok2\n");
    for (int i = 0; i < threads_n; i++) {
        paras[i].msize = 0;
        paras[i].fpout = fpout;
        paras[i].motif_indexs = malloc(sizeof(int) * 0);
    }
    //printf("ok3\n");
    for (int mi = 0; mi < motif_n; mi++) {
        int ret = rand() % threads_n;
        //printf("mi:%d, ret:%d, motif_n:%d\n", mi, ret, motif_n);
        paras[ret].motif_indexs = realloc(paras[ret].motif_indexs, sizeof(int) * (paras[ret].msize + 1));
        //printf("ok31\n");
        paras[ret].motif_indexs[paras[ret].msize] = mi;
        paras[ret].msize++;
    }
}

void free_threads(struct thread_para *paras) {
    for (int i = 0; i < threads_n; i++) {
        paras[i].msize = 0;
        free(paras[i].motif_indexs);
    }
}

void extream_gene_set(char *motif_file, char *seq_file, char *output_file) {
    FILE *fpout;
    fpout = fopen(output_file, "w");
    read_inputs(motif_file, seq_file);
    //struct thread_para *paras = malloc(sizeof(struct thread_para) * threads_n);
    struct thread_para paras[threads_n];
    assign_threads(paras, fpout);
    printf("start threading\n");
    pthread_t tids[threads_n];
    for (int i = 0; i < threads_n; i++) {
        printf("start thread %d\n", i + 1);
        int ret = pthread_create(&tids[i], NULL, extream_threading, (void *)&paras[i]);
        if(ret != 0) {
            printf("pthread_create error: %d", i);
        }
    }
    for (int i = 0; i < threads_n; i++) {
        pthread_join(tids[i], NULL);
    }
    free_threads(paras);
    //free(paras);
    fclose(fpout);
}

int main(int argc, char *argv[]) {
    if (argc == 1)
    {
        printf("Jinzer -m motif -i seq [-s seed] [-n rand_n] [-r score] -o output\n");
        exit(1);
    }
    char seq_file[max_strbuffer_len] = "seq.txt";
    char motif_file[max_strbuffer_len] = "motif.txt";
    char output_file[max_strbuffer_len] = "output.txt";
    int c;
    while((c=getopt_long(argc, argv, short_options, long_options, NULL)) != -1)
    {
        switch (c)
        {
            case 'h':
                printf("Jinzer -m motif -i seq [-s seed] [-n rand_n] [-r score] -o output\n");
                exit(0);
                break;
            case 'm':
                strcpy(motif_file, optarg);
                break;
            case 'i':
                strcpy(seq_file, optarg);
                break;
            case 's':
                random_seed = atoi(optarg);
                break;
            case 'n':
                rdm_n = atoi(optarg);
                                run_sampling = 1;
                break;
            case 'o':
                strcpy(output_file, optarg);
                break;
            case 'r':
                score_th = atof(optarg);
                break;
            default :
                printf("invalid option: %s", optarg);
                exit(1);
        }
    }
    //printf("scor_th: %f", score_th);
    srand(random_seed);
    motif_db = malloc(sizeof(struct motif) * max_motif_n);
    seq_db = malloc(sizeof(struct seq) * max_seq_n);
    if (run_sampling) {
        random_gene_set(motif_file, seq_file, output_file);
    }
    else {
        extream_gene_set(motif_file, seq_file, output_file);
    }
    free(motif_db);
    free(seq_db);
}
